from collections import OrderedDict
from hkube_python_wrapper.communication.zmq.streaming.producer.Flow import Flow
import threading


class MessageQueue(object):
    def __init__(self, consumerTypes, nodeName):
        self.nodeName = nodeName
        self.consumerTypes = consumerTypes
        self.lock = threading.Lock()
        self.indexPerConsumer = OrderedDict()
        self.sent = {}
        self.everAppended = {}
        self.lostMessages = {}
        for consumerType in self.consumerTypes:
            self.indexPerConsumer[consumerType] = 0
            self.sent[consumerType] = 0
            self.everAppended[consumerType] = 0
            self.lostMessages[consumerType] = 0
        self.sizeSum = 0
        self.queue = []

    def hasItems(self, consumerType):
        return self.indexPerConsumer[consumerType] < len(self.queue)

    def nextMessageIndex(self, consumerType):
        index = self.indexPerConsumer[consumerType]
        foundMessage = False
        while (not foundMessage) and index < len(self.queue):
            messageFlowPattern, _, _, _ = self.queue[index]
            flow = Flow(messageFlowPattern)
            if (flow.isNextInFlow(consumerType, self.nodeName)):
                foundMessage = True
            else:
                index += 1

        if (foundMessage):
            return index
        return None

    # Messages are kept in the queue until consumers of all types popped out the message.
    # An index per consumer type is maintained, to know which messages the consumer already received and conclude which message should he get now.
    def pop(self, consumerType):
        with self.lock:
            nextItemIndex = self.nextMessageIndex(consumerType)
            if (nextItemIndex is not None):
                out = self.queue[nextItemIndex]
                index = nextItemIndex + 1
                self.indexPerConsumer[consumerType] = index
                self.sent[consumerType] += 1
                while(self.removeIfNeeded()):
                    pass
                return out
            return None

    def removeIfNeeded(self):
        if (self.queue):
            anyZero = False
            out = self.queue[0]
            for consumerType, index in self.indexPerConsumer.items():
                if (index == 0):
                    messageFlowPattern, _, _, _ = out
                    flow = Flow(messageFlowPattern)
                    if (flow.isNextInFlow(consumerType, self.nodeName)):
                        anyZero = True
                        break

            if not (anyZero):
                self.queue.pop(0)
                _, _, msg, _ = out
                self.sizeSum -= len(msg)
                for consumerType in self.indexPerConsumer.keys():
                    if (self.indexPerConsumer[consumerType] > 0):
                        self.indexPerConsumer[consumerType] = self.indexPerConsumer[consumerType] - 1
                return True
        return False

    def loseMessage(self):
        with self.lock:
            if (self.queue):
                out = self.queue.pop(0)
                messageFlowPattern, _, msg, _ = out
                self.sizeSum -= len(msg)
                for consumerType in self.indexPerConsumer.keys():
                    if self.indexPerConsumer[consumerType] > 0:
                        self.indexPerConsumer[consumerType] = self.indexPerConsumer[consumerType] - 1
                    else:
                        flow = Flow(messageFlowPattern)
                        if (flow.isNextInFlow(consumerType, self.nodeName)):
                            self.lostMessages[consumerType] += 1

    # TODO: TRY TO IMPROVE THIS
    def append(self, messageFlowPattern, header, msg, appendTime):
        with self.lock:
            self.sizeSum += len(msg)
            flow = Flow(messageFlowPattern)
            hasRecipient = False
            for consumerType in self.consumerTypes:
                if (flow.isNextInFlow(consumerType, self.nodeName)):
                    self.everAppended[consumerType] += 1
                    hasRecipient = True
            if (hasRecipient):
                self.queue.append((messageFlowPattern, header, msg, appendTime))

    def size(self, consumerType):
        everAppended = self.everAppended[consumerType]
        size = everAppended - self.sent[consumerType] - self.lostMessages[consumerType]
        return size



from __future__ import print_function, division, absolute_import


class Job:
    def __init__(self, options):
        self._options = options

    @property
    def jobId(self):
        return self._options.get('jobId')

    @property
    def taskId(self):
        return self._options.get('taskId')

    @property
    def nodeName(self):
        return self._options.get('nodeName')

    @property
    def childs(self):
        return self._options.get('childs')

    @property
    def parsedFlow(self):
        return self._options.get('parsedFlow')

    @property
    def defaultFlow(self):
        return self._options.get('defaultFlow')

    @property
    def info(self):
        return self._options.get('info')

    @property
    def input(self):
        return self._options.get('input')

    @property
    def flatInput(self):
        return self._options.get('flatInput')

    @property
    def storage(self):
        return self._options.get('storage')

    @property
    def spanId(self):
        return self._options.get('spanId')

    @property
    def isStreaming(self):
        return self._options.get('kind') == 'stream'

    @property
    def isStateful(self):
        return self._options.get('stateType') != 'stateless'



import time
import zmq
import uuid
import threading
import hkube_python_wrapper.communication.zmq.streaming.signals as signals
from hkube_python_wrapper.util.logger import log

POLL_TIMEOUT_MS = 1
STOP_TIMEOUT_MS = 5000
HEARTBEAT_INTERVAL = 10
HEARTBEAT_LIVENESS_TIMEOUT = 30
INTERVAL_INIT = 1
INTERVAL_MAX = 32

context = zmq.Context()
lock = threading.Lock()

class ZMQListener(object):
    """ZMQListener
        This class runs in separate thread, but only one single thread can
        proccess data at each moment, this is dramatically decreade performance!!
    """
    def __init__(self, remoteAddress, consumerType, nodeName, onMessage, encoding, onReady=None, onNotReady=None):
        self._encoding = encoding
        self._onMessage = onMessage
        self._onReady = onReady
        self._onNotReady = onNotReady
        self._nodeName = nodeName
        self._consumerType = self._encoding.encode(consumerType, plainEncode=True)
        self._remoteAddress = remoteAddress
        self._active = True
        self._closeForce = False
        self._worker = None
        self._interval = INTERVAL_INIT
        self._lastSentTime = time.time()
        self._lastReceiveTime = time.time()
        self._ready = None
        self._readySent = False
        self._notReadySent = False
        self._isClosed = False

    def _worker_socket(self, remoteAddress):
        """Helper function that returns a new configured socket
           connected to the Paranoid Pirate queue"""
        identity = str(uuid.uuid4()).encode()
        worker = context.socket(zmq.DEALER)  # DEALER
        worker.setsockopt(zmq.IDENTITY, identity)
        worker.setsockopt(zmq.LINGER, 0)
        worker.connect(remoteAddress)
        self._send(worker, signals.PPP_INIT)
        return worker

    def _send(self, worker, signal, result=signals.PPP_EMPTY):
        if(worker):
            arr = [signal, self._consumerType, result]
            worker.send_multipart(arr, copy=False)
            self._lastSentTime = time.time()

    def _handleAMessage(self, frames):
        _, encodedMessageFlowPattern, header, message = frames  # pylint: disable=unbalanced-tuple-unpacking
        messageFlowPattern = self._encoding.decode(value=encodedMessageFlowPattern, plainEncode=True)
        return self._onMessage(messageFlowPattern, header, message)

    def start(self):  # pylint: disable=too-many-branches
        log.info("start receiving from node {node} in address {address}", node=self._nodeName, address=self._remoteAddress)
        self._worker = self._worker_socket(self._remoteAddress)

        while self._active: # pylint: disable=too-many-nested-blocks
            try:
                lockRes = lock.acquire(False)
                if(lockRes is False):
                    self._checkReady()
                    continue

                result = self._worker.poll(POLL_TIMEOUT_MS)
                if result == zmq.POLLIN:
                    frames = self._worker.recv_multipart()

                    if not frames:
                        raise Exception("Connection to producer on " + self._remoteAddress + " interrupted")

                    signal = frames[0]

                    if (signal == signals.PPP_MSG):
                        self.onNotReady()
                        result = self._handleAMessage(frames)
                        self.onReady()
                        self._send(self._worker, signals.PPP_DONE, result)

                    self._interval = INTERVAL_INIT
                    self._lastReceiveTime = time.time()
                else:
                    if (time.time() - self._lastReceiveTime > HEARTBEAT_LIVENESS_TIMEOUT):
                        self._lastReceiveTime = time.time()
                        log.warning("Heartbeat failure {addr}, Reconnecting in {interval:0.2f}", addr=self._remoteAddress, interval=self._interval)

                        if (self._interval < INTERVAL_MAX):
                            self._interval *= 2

                        if (self._active):
                            log.info("reconnecting to node {node} in address {address}", node=self._nodeName, address=self._remoteAddress)
                            self._worker.close()
                            self._worker = self._worker_socket(self._remoteAddress)

                    else:
                        self._sendHeartBeat()

            except Exception as e:
                if (self._active):
                    log.error('Error in ZMQListener {e}', e=str(e))
                    raise e
                break

            finally:
                if(lock.locked()):
                    lock.release()

        self._close()

    def _sendHeartBeat(self):
        if (time.time() - self._lastSentTime > HEARTBEAT_INTERVAL and self._active is True):
            self._send(self._worker, signals.PPP_HEARTBEAT)

    def _checkReady(self):
        if(self._ready is True and self._active is True and self._readySent is False):
            self._readySent = True
            self._notReadySent = False
            self._send(self._worker, signals.PPP_READY)
        elif (self._ready is False and self._active is True and self._notReadySent is False):
            self._notReadySent = True
            self._readySent = False
            self._send(self._worker, signals.PPP_NOT_READY)

    def ready(self):
        self._ready = True

    def notReady(self):
        self._ready = False

    def onReady(self):
        if(self._onReady and self._active):
            self._onReady(self._remoteAddress)

    def onNotReady(self):
        if(self._onNotReady and self._active):
            self._onNotReady(self._remoteAddress)

    def close(self, force=True):
        if (self._active is False):
            log.warning("Attempting to close inactive ZMQListener")
        else:
            self._active = False
            self._closeForce = force

    def waitForClose(self):
        log.info("start closing connection for node {node} in address {address}", node=self._nodeName, address=self._remoteAddress)
        while(self._isClosed is False):
            time.sleep(1)
        log.info("finish closing connection for node {node} in address {address}", node=self._nodeName, address=self._remoteAddress)

    def _close(self):
        if (self._worker is None):
            log.warning("unable to close zmqListener, worker is none")
            return

        self._send(self._worker, signals.PPP_DISCONNECT)

        if (self._closeForce is False):
            try:
                time.sleep(2)
                readAfterStopped = 0
                result = self._worker.poll(STOP_TIMEOUT_MS)
                while result == zmq.POLLIN:
                    lock.acquire()
                    try:
                        frames = self._worker.recv_multipart()
                        signal = frames[0]
                        if (signal == signals.PPP_MSG):
                            result = self._handleAMessage(frames)
                            self._send(self._worker, signals.PPP_DONE_DISCONNECT, result)
                            readAfterStopped += 1
                            log.warning('read message during stop {readAfterStopped}', readAfterStopped=readAfterStopped)
                        else:
                            log.warning('read signal {signal} during stop', signal=signal)
                    finally:
                        lock.release()
                    result = self._worker.poll(STOP_TIMEOUT_MS)
            except Exception as e:
                log.error('Error on zmqListener close {e}', e=str(e))

        self._worker.close()
        self._isClosed = True



from hkube_python_wrapper.cache.caching import Cache
import concurrent.futures
import datetime

def test_reaching_limit():
    cache = Cache({"maxCacheSize": 4})
    value1 = {"bytes": bytearray(1000000)}
    cache.update("task1", value1, 1000000)
    assert len(cache._cache) == 1
    cache.update("task2", value1, 1000000)
    cache.update("task3", value1, 1000000)
    assert len(cache._cache) == 3
    cache.update("task4", value1, 1500000)
    assert len(cache._cache) == 3
    assert "task4" in cache
    assert "task1" not in cache

def test_threading():
    cache = Cache({"maxCacheSize": 4})
    # fill cache
    startTime=datetime.datetime.now()
    for i in range(10000):
        values=(str(i),{"bytes": bytearray(500)},500)
        cache._cache[values[0]] = {'timestamp': startTime+datetime.timedelta(0,1), 'size': values[2], 'value': values[1], 'header': None}
    request=[]
    for i in range(1000):
        request.append(('key'+str(i),{"bytes": bytearray(50000)},50000))
    res=[]
    def process(item):
        cache.update(item[0], item[1])
        return item[0]
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        for out in executor.map(process,request):
            res.append(out)
    assert len(res) == 1000

def test_too_large_message():
    cache = Cache({"maxCacheSize": 5})
    value1 = {"bytes": bytearray(1000000)}
    value2 = {"bytes": bytearray(5000000)}
    result = cache.update("task1", value1, 1000000)
    assert result == True
    assert len(cache._cache) == 1
    result = cache.update("task2", value2, 6000000)
    assert result == False
    assert len(cache._cache) == 1


def test_get_all():
    cache = Cache({"maxCacheSize": 5})
    value1 = {"data": "1"}
    value2 = {"data": "2"}
    result = cache.update("task1", value1)
    assert result == True
    assert len(cache._cache) == 1
    cache.update("task2", value2)
    assert len(cache._cache) == 2
    tasksNotInCache, valuesInCache = cache.getAll(['task1', 'task2', 'task3', 'task4'])
    assert len(tasksNotInCache) == 2
    assert "task3" in set(tasksNotInCache)
    assert "task4" in set(tasksNotInCache)
    assert len(valuesInCache) == 2
    data = list(map(lambda value: value.get('data'), valuesInCache))
    assert "1" in set(data)
    assert "2" in set(data)

if __name__ == '__main__':
    test_threading()


#
# Based on Paranoid Pirate queue
#
#   Author: Daniel Lundin <dln(at)eintr(dot)org>
#
import time
from .Flow import Flow
from .WorkerQueue import WorkerQueue
from .MessageQueue import MessageQueue
from hkube_python_wrapper.util.logger import log
import hkube_python_wrapper.communication.zmq.streaming.signals as signals
import zmq

HEARTBEAT_INTERVAL = 10
HEARTBEAT_LIVENESS = 5
CYCLE_LENGTH_MS = 1
PURGE_INTERVAL = 5

context = zmq.Context()

class ZMQProducer(object):
    def __init__(self, port, maxMemorySize, responseAccumulator, queueTimeAccumulator, consumerTypes, encoding, nodeName):
        self.nodeName = nodeName
        self.encoding = encoding
        self.responseAccumulator = responseAccumulator
        self.queueTimeAccumulator = queueTimeAccumulator
        self.maxMemorySize = maxMemorySize
        self.port = port
        self.consumerTypes = consumerTypes
        self.messageQueue = MessageQueue(consumerTypes, self.nodeName)
        self._backend = context.socket(zmq.ROUTER)  # ROUTER
        self._backend.bind("tcp://*:" + str(port))  # For workers
        log.info("Producer listening on tcp://*:{port}", port=port)
        self.active = True
        self.watingForResponse = {}
        self._nextPurgeTime = time.time() + PURGE_INTERVAL
        self._nextHeartbeat = time.time() + HEARTBEAT_INTERVAL

    def produce(self, header, message, messageFlowPattern=[]):
        appendTime = time.time()
        while (self.messageQueue.sizeSum > self.maxMemorySize):
            self.messageQueue.loseMessage()
        self.messageQueue.append(messageFlowPattern, header, message, appendTime)

    def start(self):  # pylint: disable=too-many-branches
        poll_workers = zmq.Poller()
        poll_workers.register(self._backend, zmq.POLLIN)
        workers = WorkerQueue(self.consumerTypes)

        while self.active:  # pylint: disable=too-many-nested-blocks
            try:
                socks = dict(poll_workers.poll(CYCLE_LENGTH_MS))

                if socks.get(self._backend) == zmq.POLLIN:

                    frames = self._backend.recv_multipart() or []

                    if(len(frames) != 4):
                        log.warning("got {len} frames {frames}", len=len(frames), frames=frames)
                        continue

                    address, signal, consumer, result = frames # pylint: disable=unbalanced-tuple-unpacking
                    consumerType = self.encoding.decode(value=consumer, plainEncode=True)

                    if (not consumerType in self.consumerTypes):
                        log.warning("Producer got message from unknown consumer: {consumerType}, dropping the message", consumerType=consumerType)
                        continue

                    if (signal in (signals.PPP_NOT_READY, signals.PPP_DISCONNECT)):
                        workers.notReady(consumerType, address)
                        continue

                    if (signal in (signals.PPP_DONE, signals.PPP_DONE_DISCONNECT)):
                        sentTime = self.watingForResponse.get(address)
                        if (sentTime):
                            del self.watingForResponse[address]
                            roundTripTime = round((time.time() - sentTime) * 1000, 4)
                            self.responseAccumulator(result, consumerType, roundTripTime)
                        else:
                            log.error('missing from watingForResponse:' + str(signal))

                    if (address not in self.watingForResponse and signal in (signals.PPP_INIT, signals.PPP_READY, signals.PPP_HEARTBEAT, signals.PPP_DONE)):
                        workers.ready(consumerType, address)

                    if (time.time() >= self._nextHeartbeat):
                        for workersOfType in workers.queues.values():
                            for worker in workersOfType.values():
                                self._send([worker.address, signals.PPP_HEARTBEAT])

                        self._nextHeartbeat = time.time() + HEARTBEAT_INTERVAL

                for consumerType, workerQueue in workers.queues.items():
                    if (workerQueue):
                        message = self.messageQueue.pop(consumerType)
                        if (message):
                            messageFlowPattern, header, payload, appendTime = message
                            worker = workers.nextWorker(consumerType)
                            flow = Flow(messageFlowPattern)
                            flowMsg = self.encoding.encode(flow.getRestOfFlow(self.nodeName), plainEncode=True)
                            frames = [worker, signals.PPP_MSG, flowMsg, header, payload]
                            self.watingForResponse[worker] = time.time()
                            queueTime = round((time.time() - appendTime) * 1000, 4)
                            self.queueTimeAccumulator(consumerType, queueTime)
                            self._send(frames)

            except Exception as e:
                if (self.active):
                    log.error('Error in ZMQProducer {e}', e=str(e))
                else:
                    break

            if(time.time() > self._nextPurgeTime):
                self._nextPurgeTime = time.time() + PURGE_INTERVAL
                workers.purge()

    def _send(self, frames):
        self._backend.send_multipart(frames, copy=False)

    def queueSize(self, consumerSize):
        return self.messageQueue.size(consumerSize)

    def sent(self, consumerType):
        return self.messageQueue.sent[consumerType]

    def close(self, force=True):
        log.info('queue size before close = {len}', len=len(self.messageQueue.queue))
        while self.messageQueue.queue and not force:
            log.info('queue size during close = {len}', len=len(self.messageQueue.queue))
            time.sleep(1)
        log.info('queue size after close = {len}', len=len(self.messageQueue.queue))
        self._backend.close()
        self.active = False





PPP_INIT = b"\x01"  # Signals worker is ready
PPP_HEARTBEAT = b"\x02"  # Signals worker heartbeat
PPP_DISCONNECT = b"\x03"  # Disconnect
PPP_READY = b"\x04"  # Signals worker is not ready
PPP_NOT_READY = b"\x05"  # Signals worker is not ready
PPP_DONE = b"\x06"
PPP_EMPTY = b"\x07"
PPP_MSG = b"\x08"
PPP_DONE_DISCONNECT = b"\x09"



from collections import OrderedDict
import time
from .Worker import Worker
from random import randrange
from hkube_python_wrapper.util.logger import log


class WorkerQueue(object):
    def __init__(self, consumerTypes):
        self.queues = {}
        for consumerType in consumerTypes:
            self.queues[consumerType] = OrderedDict()

    def ready(self, consumerType, address):
        worker = Worker(address)
        self.queues[consumerType].pop(address, None)
        self.queues[consumerType][address] = worker

    def notReady(self, consumerType, address):
        self.queues[consumerType].pop(address, None)

    def purge(self):
        """Look for & kill expired workers."""
        t = time.time()
        expired = []
        for consumer, queue in self.queues.items():
            for address, worker in queue.items():
                if t > worker.expiry:  # Worker expired
                    expired.append((address, consumer))
            for (address, consumerType) in expired:
                log.warning("Idle worker expired: {address}", address=address)
                self.queues[consumerType].pop(address, None)

    def nextWorker(self, consumerType):
        workers = list(self.queues[consumerType].keys())
        rand = randrange(len(workers))
        worker = workers[rand]
        self.queues[consumerType].pop(worker, None)
        return worker
